package ctbrec.sites.cherrytv;

import com.iheartradio.m3u8.*;
import com.iheartradio.m3u8.data.MasterPlaylist;
import com.iheartradio.m3u8.data.Playlist;
import com.iheartradio.m3u8.data.PlaylistData;
import ctbrec.AbstractModel;
import ctbrec.Config;
import ctbrec.NotImplementedExcetion;
import ctbrec.io.HttpException;
import ctbrec.recorder.download.StreamSource;
import okhttp3.Request;
import okhttp3.Response;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;
import java.util.concurrent.ExecutionException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static ctbrec.Model.State.*;
import static ctbrec.io.HttpConstants.*;
import static java.nio.charset.StandardCharsets.UTF_8;

public class CherryTvModel extends AbstractModel {

    private static final Logger LOG = LoggerFactory.getLogger(CherryTvModel.class);
    private static final Pattern NEXT_DATA = Pattern.compile("<script id=\"__NEXT_DATA__\" type=\"application/json\">(.*?)</script>");

    private boolean online = false;
    private int[] resolution;
    private String masterPlaylistUrl;

    @Override
    public boolean isOnline(boolean ignoreCache) throws IOException, ExecutionException, InterruptedException {
        if (ignoreCache) {
            String url = getUrl();
            Request req = new Request.Builder().url(url)
                    .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                    .header(ACCEPT, "*")
                    .header(ACCEPT_LANGUAGE, "en")
                    .header(REFERER, getSite().getBaseUrl())
                    .build();
            try (Response resp = site.getHttpClient().execute(req)) {
                String body = Objects.requireNonNull(resp.body()).string();
                Files.write(Paths.get("/tmp/mdl.html"), body.getBytes(StandardCharsets.UTF_8));
                Matcher m = NEXT_DATA.matcher(body);
                if (m.find()) {
                    JSONObject json = new JSONObject(m.group(1));
                    JSONObject apolloState = json.getJSONObject("props").getJSONObject("pageProps").getJSONObject("apolloState");
                    online = false;
                    onlineState = OFFLINE;
                    for (Iterator<String> iter = apolloState.keys(); iter.hasNext();) {
                        String key = iter.next();
                        if (key.startsWith("Broadcast:")) {
                            JSONObject broadcast = apolloState.getJSONObject(key);
                            setDisplayName(broadcast.optString("title"));
                            // id = broadcast.getString("id");
                            // roomId = broadcast.getString("roomId");
                            online = broadcast.optString("showStatus").equalsIgnoreCase("Public")
                                    && broadcast.optString("broadcastStatus").equalsIgnoreCase("Live");
                            onlineState = online ? ONLINE : OFFLINE;
                            masterPlaylistUrl = broadcast.optString("pullUrl", null);
                            break;
                        }
                    }
                } else {
                    LOG.error("NEXT_DATA not found in model page {}", getUrl());
                    return false;
                }
            } catch (JSONException e) {
                LOG.error("Unable to determine online state for {}. Probably the JSON structure in NEXT_DATA changed", getName());
            }
        }
        return online;
    }

    public void setOnline(boolean online) {
        this.online = online;
    }

    @Override
    public State getOnlineState(boolean failFast) throws IOException, ExecutionException {
        if (!failFast) {
            try {
                isOnline(true);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                onlineState = OFFLINE;
            } catch (IOException | ExecutionException e) {
                onlineState = OFFLINE;
            }
        }
        return onlineState;
    }

    @Override
    public void setOnlineState(State onlineState) {
        this.onlineState = onlineState;
    }

    @Override
    public List<StreamSource> getStreamSources() throws IOException, ExecutionException, ParseException, PlaylistException {
        try {
            isOnline(true);
            MasterPlaylist masterPlaylist = getMasterPlaylist();
            List<StreamSource> sources = new ArrayList<>();
            for (PlaylistData playlist : masterPlaylist.getPlaylists()) {
                if (playlist.hasStreamInfo()) {
                    StreamSource src = new StreamSource();
                    src.bandwidth = playlist.getStreamInfo().getBandwidth();
                    src.height = playlist.getStreamInfo().getResolution().height;
                    String masterUrl = masterPlaylistUrl;
                    String baseUrl = masterUrl.substring(0, masterUrl.lastIndexOf('/') + 1);
                    String segmentUri = baseUrl + playlist.getUri();
                    src.mediaPlaylistUrl = segmentUri;
                    if(src.mediaPlaylistUrl.contains("?")) {
                        src.mediaPlaylistUrl = src.mediaPlaylistUrl.substring(0, src.mediaPlaylistUrl.lastIndexOf('?'));
                    }
                    LOG.trace("Media playlist {}", src.mediaPlaylistUrl);
                    sources.add(src);
                }
            }
            return sources;
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new ExecutionException(e);
        }
    }

    private MasterPlaylist getMasterPlaylist() throws IOException, ParseException, PlaylistException {
        LOG.trace("Loading master playlist {}", masterPlaylistUrl);
        Request req = new Request.Builder()
                .url(masterPlaylistUrl)
                .header(USER_AGENT, Config.getInstance().getSettings().httpUserAgent)
                .build();
        try (Response response = getSite().getHttpClient().execute(req)) {
            if (response.isSuccessful()) {
                String body = Objects.requireNonNull(response.body()).string();
                LOG.trace(body);
                InputStream inputStream = new ByteArrayInputStream(body.getBytes(UTF_8));
                PlaylistParser parser = new PlaylistParser(inputStream, Format.EXT_M3U, Encoding.UTF_8, ParsingMode.LENIENT);
                Playlist playlist = parser.parse();
                MasterPlaylist master = playlist.getMasterPlaylist();
                return master;
            } else {
                throw new HttpException(response.code(), response.message());
            }
        }
    }

    @Override
    public void invalidateCacheEntries() {
        resolution = null;
    }

    @Override
    public void receiveTip(Double tokens) throws IOException {
        throw new NotImplementedExcetion();
    }

    @Override
    public int[] getStreamResolution(boolean failFast) throws ExecutionException {
        if(resolution == null) {
            if(failFast) {
                return new int[2];
            }
            try {
                if(!isOnline()) {
                    return new int[2];
                }
                List<StreamSource> sources = getStreamSources();
                Collections.sort(sources);
                StreamSource best = sources.get(sources.size()-1);
                resolution = new int[] {best.width, best.height};
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                LOG.warn("Couldn't determine stream resolution for {} - {}", getName(), e.getMessage());
                resolution = new int[2];
            } catch (ExecutionException | IOException | ParseException | PlaylistException e) {
                LOG.warn("Couldn't determine stream resolution for {} - {}", getName(), e.getMessage());
                resolution = new int[2];
            }
            return resolution;
        } else {
            return resolution;
        }
    }

    @Override
    public boolean follow() throws IOException {
        return false;
    }

    @Override
    public boolean unfollow() throws IOException {
        return false;
    }

    public void mapOnlineState(String roomState) {
        switch (roomState) {
        case "private":
        case "fullprivate":
            setOnlineState(PRIVATE);
            break;
        case "group":
        case "public":
            setOnlineState(ONLINE);
            setOnline(true);
            break;
        default:
            LOG.debug(roomState);
            setOnlineState(OFFLINE);
        }
    }
}
